/*part of mspgcc, modified file from TI:
- Removed flash writing code as this does not work over the parallel port
  under a non realtime OS.
- Also removed BlowFuse code as it's not supported by the FET kit.
- Fusecheck delay inserted.
- added some debug outputs, see def.h for activation.

chris

original TI notes:
*/
/*==========================================================================*\
|                                                                            |
| JTAGfunc.c                                                                 |
|                                                                            |
| JTAG Control Sequences for Erasing / Programming / Fuse Burning            |
|----------------------------------------------------------------------------|
| Project:              JTAG Functions                                       |
| Developed using:      IAR Embedded Workbench 2.31C                         |
|----------------------------------------------------------------------------|
| Author:               FRGR                                                 |
| Version:              1.2                                                  |
| Initial Version:      04-17-02                                             |
| Last Change:          08-29-02                                             |
|----------------------------------------------------------------------------|
| Version history:                                                           |
| 1.0 04/02 FRGR        Initial version.                                     |
| 1.1 04/02 ALB2        Formatting changes, added comments.                  |
| 1.2 08/02 ALB2        Initial code release with Lit# SLAA149.              |
|----------------------------------------------------------------------------|
| Designed 2002 by Texas Instruments Germany                                 |
\*==========================================================================*/

#include "defs.h"
#include "HIL.h"
#include "JTAGfunc.h"
#include "funclets.h"   //syncCPU


WORD DEVICE = 0;        // Global target device ID (e.g. 0xF149); is set 
                        // during GetDevice() function execution.

//----------------------------------------------------------------------------
/* Reset target JTAG interface and perform fuse-HW check.
   Arguments: None
   Result:    None
*/
void ResetTAP(void) {
    WORD i;
    
    // Perform fuse check
#ifdef SLOWFUSECHECK_BUG // Silicon bug requires slow check of fuse.
    HIL_TCK(0);
    HIL_TMS(0);
    HIL_DelayMSec(FUSECHECK_DELAY);
    HIL_TMS(1);
    HIL_DelayMSec(FUSECHECK_DELAY);
    HIL_TMS(0);
    HIL_DelayMSec(FUSECHECK_DELAY);
    HIL_TMS(1);
    HIL_DelayMSec(FUSECHECK_DELAY);
#else
    HIL_TMS(0);
    HIL_TMS(1);
    HIL_TMS(0);
    HIL_TMS(1);
#endif
    // Now fuse is checked, Reset JTAG FSM
    for (i = 6; i > 0; i--) {
        HIL_TCK(POS_EDGE);
    }
    // JTAG FSM is now in Test-Logic-Reset  
    ClrTCK();
    ClrTMS();
    SetTCK();
    // JTAG FSM is now in Run-Test/IDLE
}

//----------------------------------------------------------------------------
/* Function to execute a Power-Up Clear (PUC) using JTAG CNTRL SIG register
   keep JTAG connection. The watchdog is disabled and the executionis stopped.
   Arguments: None
   Result:    WORD (STATUS_OK if JTAG ID is valid, STATUS_ERROR otherwise)
*/
WORD ExecutePUC(void) {
    WORD JTAGVERSION;

    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2C01, F_WORD);                        // Apply Reset
    HIL_JTAG_DR(0x2401, F_WORD);                        // Remove Reset
    HIL_TCLK(POS_EDGE);
    HIL_TCLK(POS_EDGE);
    ClrTCLK();
    JTAGVERSION = HIL_JTAG_IR(IR_ADDR_CAPTURE);
    SetTCLK();
    
    WriteMem(F_WORD, 0x0120, 0x5A80);                   // Disable Watchdog on target device  
    
    if (JTAGVERSION != JTAG_ID) {
        MSP430_Log(1, "JTAGfunc: JTAG ID wrong.\n");     //DEBUG
        return(STATUS_ERROR);
    }
    
    return(STATUS_OK);
}

//----------------------------------------------------------------------------
/* Function to take target device under JTAG control. Disables the target
   watchdog. Sets the global DEVICE variable as read from the target device.
   Arguments: None
   Result:    WORD (STATUS_ERROR if fuse is blown, incorrect JTAG ID or
                    synchronizing time-out; STATUS_OK otherwise)
*/
WORD GetDevice(void) {
    WORD i;
    WORD ctrl;
    
    DEVICE = 0;                                         // Preset DEVICE with "not a device"
    ResetTAP();                                         // Reset JTAG state machine, check fuse HW
   
    if (IsFuseBlown()) {                                // Stop here if fuse is already blown
        return(STATUS_ERROR);
    }

    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2401, F_WORD);                        // Set device into JTAG mode + read
    if (HIL_JTAG_IR(IR_CNTRL_SIG_CAPTURE) != JTAG_ID) {
        MSP430_Log(1, "JTAGfunc: JTAG ID wrong\n");     //DEBUG
        return(STATUS_ERROR);
    }
    
    // Wait until CPU is synchronized, timeout after a limited # of attempts
    for (i = 50; i > 0; i--) {
        ctrl = HIL_JTAG_DR(0x0000, F_WORD);
        MSP430_Log(6, "JTAGfunc: JTAG CNTRL: %04x\n", ctrl);
        if (ctrl & CNTRL_SIG_TCE) {
            DEVICE = ReadMem(F_WORD, 0x0FF0);           // Get target device type 
                                                        //(bytes are interchanged)
            DEVICE = (DEVICE << 8) + (DEVICE >> 8);     // Set global DEVICE type 
            MSP430_Log(1, "JTAGfunc: Sync OK, device: 0x%04x\n", DEVICE);   //DEBUG
            break;
        } else if (i == 1) {
            MSP430_Log(1, "JTAGfunc: Sync failed\n");   //DEBUG
            return(STATUS_ERROR);                       // Timeout reached, return false
        } else {
            HIL_DelayMSec(1);
        }
    }
    if (ExecutePUC() != STATUS_OK) {                    // Perform PUC, Includes  
        MSP430_Log(1, "JTAGfunc: PUC failed\n");        //DEBUG
        return(STATUS_ERROR);                           // target Watchdog disable.
    }
    if (syncCPU() != STATUS_OK) {                       // verify state
        MSP430_Log(1, "JTAGfunc: syncCPU failed\n");        //DEBUG
        return(STATUS_ERROR);                           // target Watchdog disable.
    }
    
    return(STATUS_OK);
}

//----------------------------------------------------------------------------
/* Function to set target CPU JTAG FSM into the instruction fetch state
   Argument: None
   Result:   WORD (STATUS_OK if instr. fetch was set, STATUS_ERROR otherwise)
*/
WORD SetInstrFetch(void) {
    WORD i;
    
    HIL_JTAG_IR(IR_CNTRL_SIG_CAPTURE);
    
    // Wait until CPU is in instr. fetch state, timeout after limited attempts
    for (i = 50; i > 0; i--) {
        HIL_TCLK(POS_EDGE);
        if ((HIL_JTAG_DR(0x0000, F_WORD) & 0x0080) == 0x0080) {
            return(STATUS_OK);
        }
    }
    return(STATUS_ERROR); 
}

//----------------------------------------------------------------------------
/* Load a given address into the target CPU's program counter (PC).
   Argument: WORD Addr (destination address)
   Result:   None
*/
WORD SetPC(WORD Addr) {
    SetInstrFetch();                            // Set CPU into instruction fetch mode, TCLK=1
    
    // Load PC with address
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x3401, F_WORD);                // CPU has control of RW & BYTE.
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(MOV_IMM_PC, F_WORD);            // "mov #addr,PC" instruction
    HIL_TCLK(NEG_EDGE);
    HIL_JTAG_DR(Addr, F_WORD);                  // Send addr value
    HIL_TCLK(NEG_EDGE);
    HIL_TCLK(NEG_EDGE);                         // Now the PC should be on Addr
    HIL_JTAG_IR(IR_ADDR_CAPTURE);
    if (Addr != HIL_JTAG_DR(0, 16)) return STATUS_ERROR;//ERROR: SetPC failed!
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2401, F_WORD);                // JTAG has control of RW & BYTE.
    return STATUS_OK;
}

//----------------------------------------------------------------------------
/* Load a given address into the target CPU's program counter (PC).
   Argument: WORD Addr (destination address)
   Result:   None
*/
WORD SetReg(BYTE Regnum, WORD Value) {
    // Verify parameters
    if (Regnum > 15) {
        return STATUS_ERROR;
    }
    
    SetInstrFetch();                            // Set CPU into instruction fetch mode, TCLK=1
    
    // Load register with value
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x3401, F_WORD);                // CPU has control of RW & BYTE.
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(MOV_IMM_RX | Regnum, F_WORD); // "mov #addr, Rx" instruction
    HIL_TCLK(NEG_EDGE);                         // instr fetch
    HIL_JTAG_DR(Value, F_WORD);                 // Send addr value
    HIL_TCLK(NEG_EDGE);                         // data fetch
    HIL_TCLK(NEG_EDGE);                         // instr exec
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2401, F_WORD);                // JTAG has control of RW & BYTE.
    return STATUS_OK;
}

//----------------------------------------------------------------------------
/* Load a given address into the target CPU's program counter (PC).
   Argument: WORD Addr (destination address)
   Result:   None
*/
WORD GetReg(BYTE Regnum, WORD *Value) {
    // Verify parameters
    if (Regnum > 15) {
        return STATUS_ERROR;
    }
    
    SetInstrFetch();                            // Set CPU into instruction fetch mode, TCLK=1
    
    // Load PC with address
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x3401, F_WORD);                // CPU has control of RW & BYTE.
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(MOV_RX_MEM | (Regnum << 8), F_WORD); //"mov Rx, &a"
    HIL_TCLK(NEG_EDGE);                         // instr fetch
    HIL_JTAG_DR(SAVE_ADDRESS, F_WORD);          // Dummy addr value
    HIL_TCLK(NEG_EDGE);                         // data fetch
    HIL_TCLK(NEG_EDGE);                         // instr exec
    HIL_TCLK(NEG_EDGE);                         // (writes need 2 clks)
    HIL_JTAG_IR(IR_DATA_CAPTURE);
    *Value = HIL_JTAG_DR(0, F_WORD);            // Read databus which contains the regsiters value
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2401, F_WORD);                // JTAG has control of RW & BYTE.
    return STATUS_OK;
}

//----------------------------------------------------------------------------
/* Function to set the CPU into a controlled stop state
*/
void HaltCPU(void) {
    SetInstrFetch();                            // Set CPU into instruction fetch mode
    
    ClrTCLK();                                  //lch
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2401, F_WORD);
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(0x3FFF, F_WORD);                // Send JMP $ instruction
    HIL_TCLK(NEG_EDGE);
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2409, F_WORD);                // Set JTAG_HALT bit
    SetTCLK();
}

//----------------------------------------------------------------------------
/* Function to release the target CPU from the controlled stop state
*/
void ReleaseCPU(void) {
    ClrTCLK();
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2401, F_WORD);                // Clear the HALT_JTAG bit
    HIL_JTAG_IR(IR_ADDR_CAPTURE);
    SetTCLK();
}

//----------------------------------------------------------------------------
/* Function to release the target device from JTAG control
   Argument: WORD Addr (0xFFFE: Perform Reset, means Load Reset Vector into PC,
                        otherwise: Load Addr into PC)
   Result:   None
*/
void ReleaseDevice(WORD Addr) {
    if (Addr == V_RESET) {
        HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
        HIL_JTAG_DR(0x2C01, F_WORD);            // Perform a PUC reset
        HIL_JTAG_DR(0x2401, F_WORD);
    } else {
        SetPC(Addr);                            // Set target CPU's PC
    }
    HIL_JTAG_IR(IR_CNTRL_SIG_RELEASE);
}

//----------------------------------------------------------------------------
/* This function compares the computed PSA (Pseudo Signature Analysis) value
   to the PSA value shifted out from the target device.
   It is used for very fast data block write or erasure verification.
   Arguments: WORD StartAddr (Start address of data block to be checked)
              WORD Length (Number of WORDs within data block)
              WORD *DataArray (Pointer to array with the data, 0 for Erase Check)
   Result:    WORD (STATUS_OK if comparison was successful, STATUS_ERROR otherwise)
*/
WORD VerifyPSA(WORD StartAddr, WORD Length, WORD *DataArray) {
    WORD TDOword, i;
    WORD POLY = 0x0805;                         // Polynom value for PSA calculation
    WORD PSA_CRC = StartAddr-2;                 // Start value for PSA calculation
    
    ExecutePUC();          
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2401, F_WORD);
    SetInstrFetch();
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(0x4030, F_WORD);
    HIL_TCLK(NEG_EDGE);
    HIL_JTAG_DR(StartAddr-2, F_WORD);
    HIL_TCLK(NEG_EDGE);
    HIL_TCLK(NEG_EDGE);
    HIL_TCLK(NEG_EDGE);
    HIL_JTAG_IR(IR_ADDR_CAPTURE);
    HIL_JTAG_DR(0x0000, F_WORD);
    HIL_JTAG_IR(IR_DATA_PSA);
    for (i = 0; i < Length; i++) {
        // Calculate the PSA (Pseudo Signature Analysis) value  
        if ((PSA_CRC & 0x8000) == 0x8000) {
            PSA_CRC ^= POLY;
            PSA_CRC <<= 1;
            PSA_CRC |= 0x0001;
        } else {
            PSA_CRC <<= 1;
        }
        // if pointer is 0 then use erase check mask, otherwise data  
        &DataArray[0] == 0 ? (PSA_CRC ^= 0xFFFF) : (PSA_CRC ^= DataArray[i]);
        
        // Clock through the PSA  
        SetTCLK();
        ClrTCK();
        SetTMS();
        SetTCK();                               // Select DR scan
        ClrTCK();
        ClrTMS();
        SetTCK();                               // Capture DR
        ClrTCK();
        SetTCK();                               // Shift DR
        ClrTCK();
        SetTMS();
        SetTCK();                               // Exit DR
        ClrTCK();
        SetTCK();
        ClrTMS();
        ClrTCK();
        SetTCK();
        ClrTCLK();
    }
    HIL_JTAG_IR(IR_SHIFT_OUT_PSA);
    TDOword = HIL_JTAG_DR(0x0000, F_WORD);      // Read out the PSA value
    SetTCLK();
    
    return((TDOword == PSA_CRC) ? STATUS_OK : STATUS_ERROR);
}  

//----------------------------------------------------------------------------
/* This function writes one byte/WORD at a given address ( <0xA00) 
   Arguments: WORD Format (F_BYTE or F_WORD)
              WORD Addr (Address of data to be written)
              WORD Data (shifted data)  
   Result:    None
*/
void WriteMem(WORD Format, WORD Addr, WORD Data) {
    HaltCPU();
    
    ClrTCLK();
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    if (Format == F_WORD) {
        HIL_JTAG_DR(0x2408, F_WORD);            // Set word write
    } else {
        HIL_JTAG_DR(0x2418, F_WORD);            // Set byte write
    }
    HIL_JTAG_IR(IR_ADDR_16BIT);
    HIL_JTAG_DR(Addr, F_WORD);                  // Set addr
    HIL_JTAG_IR(IR_DATA_TO_ADDR);
    HIL_JTAG_DR(Data, F_WORD);                  // Shift in 16 bits
    SetTCLK();
    
    ReleaseCPU();
}

//----------------------------------------------------------------------------
/* This function writes an array of words into the target memory.
   Arguments: WORD StartAddr (Start address of target memory)
              WORD Length (Number of words to be programmed)
              WORD *DataArray (Pointer to array with the data)
   Result:    None
*/
void WriteMemQuick(WORD StartAddr, WORD Length, const WORD *DataArray) {
    WORD i;
    
    // Initialize writing:
    SetPC((WORD)(StartAddr-4));
    HaltCPU();
    
    ClrTCLK();
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2408, F_WORD);                // Set RW to write
    HIL_JTAG_IR(IR_DATA_QUICK);
    for (i = 0; i < Length; i++) {
        HIL_JTAG_DR(DataArray[i], F_WORD);      // Shift in the write data
        SetTCLK();
        ClrTCLK();                              // Increment PC by 2
    }
    SetTCLK();
    
    ReleaseCPU();
}

//----------------------------------------------------------------------------
/* This function reads one byte/word from a given address in memory
   Arguments: WORD Format (F_BYTE or F_WORD)
              WORD Addr (address of memory)
   Result:    WORD (content of the addressed memory location)
*/
WORD ReadMem(WORD Format, WORD Addr) {
    WORD TDOword;
    
    HaltCPU();
    
    ClrTCLK();
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    if  (Format == F_WORD) {
        HIL_JTAG_DR(0x2409, F_WORD);            // Set word read
    } else {
        HIL_JTAG_DR(0x2419, F_WORD);            // Set byte read
    }
    HIL_JTAG_IR(IR_ADDR_16BIT);
    HIL_JTAG_DR(Addr, F_WORD);                  // Set address
    HIL_JTAG_IR(IR_DATA_TO_ADDR);
    SetTCLK();
    
    ClrTCLK();
    TDOword = HIL_JTAG_DR(0x0000, F_WORD);      // Shift out 16 bits
    SetTCLK();
    
    ReleaseCPU();
    return(Format == F_WORD ? TDOword : TDOword & 0x00FF);
}

//----------------------------------------------------------------------------
/* This function reads an array of words from a memory.
   Arguments: WORD StartAddr (Start address of memory to be read)
              WORD Length (Number of words to be read)
              WORD *DataArray (Pointer to array for the data)
   Result:    None
*/
void ReadMemQuick(WORD StartAddr, WORD Length, WORD *DataArray) {
    WORD i;
    
    // Initialize reading:
    SetPC(StartAddr-4);         //XXX: is this correct or should it be -2
    HaltCPU();
    
    ClrTCLK();
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2409, F_WORD);                    // Set RW to read
    HIL_JTAG_IR(IR_DATA_QUICK);
    
    for (i = 0; i < Length; i++) {
        SetTCLK();
        ClrTCLK();
        DataArray[i] = HIL_JTAG_DR(0x0000, F_WORD); // Shift out the data
                                                    // from the target.
    }
    SetTCLK();
    
    ReleaseCPU();
}

//----------------------------------------------------------------------------
/* This function performs an Erase Check over the given memory range
   Arguments: WORD StartAddr (Start address of memory to be checked)
              WORD Length (Number of words to be checked)
   Result:    WORD (STATUS_OK if erase check was successful, STATUS_ERROR otherwise)
*/
WORD EraseCheck(WORD StartAddr, WORD Length) {
    return (VerifyPSA(StartAddr, Length, 0));
}

//----------------------------------------------------------------------------
/* This function performs a Verification over the given memory range
   Arguments: WORD StartAddr (Start address of memory to be verified)
              WORD Length (Number of words to be verified)
              WORD *DataArray (Pointer to array with the data)
   Result:    WORD (STATUS_OK if verification was successful, STATUS_ERROR otherwise)
*/
WORD VerifyMem(WORD StartAddr, WORD Length, WORD *DataArray) {
    return (VerifyPSA(StartAddr, Length, &DataArray[0]));
}


//------------------------------------------------------------------------
/* This function checks if the JTAG access security fuse is blown.
   Arguments: None
   Result:    WORD (1 if fuse is blown, 0 otherwise)
*/
WORD IsFuseBlown(void) {
    WORD i;
    WORD answer;
  
    for (i = 3; i > 0; i--) {                   // First trial could be wrong
        HIL_JTAG_IR(IR_CNTRL_SIG_CAPTURE);
        if (HIL_JTAG_DR(0xAAAA, F_WORD) == 0x5555) {
            MSP430_Log(1, "JTAGfunc: JTAG fuse burned\n"); //DEBUG
            return 1;                           // Fuse is blown
        }
    }
    //try to find out more details...
    answer = HIL_JTAG_DR(0, F_WORD);
    if ((answer != 0) && (answer != 0xffff)) {
        MSP430_Log(1, "JTAGfunc: JTAG fuse ok\n"); //DEBUG
    } else {
        MSP430_Log(1, "JTAGfunc: ERROR: possibly no device\n"); //DEBUG
    }
    return 0;                                   // fuse is not blown
}


/****************************************************************************/
/*                         END OF SOURCE FILE                               */
/****************************************************************************/
